{
    
    LNK Library for Cobra
    Analyses the properties of windows shortcut (.lnk) files in depth,
    and provides an interface to create new shortcuts.    
    Version 3, 20080701

    ---------------------------------------    
       
    Copyright (c) 2008 Ben Golightly

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following
    conditions:
    
    Except as contained in this notice, the name(s) of the above
    copyright holders shall not be used in advertising or otherwise
    to promote the sale, use or other dealings in this Software
    without prior written authorization.
    
    The end-user documentation included with the redistribution, if
    any, must include the following acknowledgment: "This product
    includes software developed by TopHat Stuff and its contributors",
    in the same place and form as other third-party acknowledgments.
    Alternately, this acknowledgment may appear in the software itself,
    in the same form and location as other such third-party
    acknowledgments.

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.
    
    ---------------------------------------
    
    Credits:
    
    I've spent a lot of time examining different files in hex editors
    to create this unit, but I must thank the following source for the
    invaluable help it provided:
        
        The Windows Shortcut File Format as reverse-engineered by Jesse Hager
        Document Version 1.0
        http://mediasrv.ns.ac.yu/extra/fileformat/windows/lnk/shortcut.pdf
    
    However, as a note to other people who may follow that guide, there were
    some very important omissions that I have noted in my code as very important.
    
    Also, XVI is a great hex editor for windows:
    
        Freeware Hex Editor XVI32
        http://www.chmaas.handshake.de/delphi/freeware/xvi32/xvi32.htm
    
    ---------------------------------------
    
    Exported functions and procedures:
        
        Function ReadShortcut(lnkpath: String) : ^lnks
        Function CreateShortcut(lnkpath: String) : ^lnks
        Function WriteShortcut(lnk: ^lnks) : Boolean
    
}

Unit

Const
        
    // Flag Properties
    LNK_SHELL_PRESENT               = 0 ; Export    // The shell item id list is present
    LNK_POINTS_TO_FILE_OR_DIR       = 1 ; Export    // Points to a file or directory
    LNK_HAS_DESCRIPTION             = 2 ; Export    // Has a description string
    LNK_HAS_RELATIVE_PATH           = 3 ; Export    // Has a relative path string
    LNK_HAS_WORKING_DIR             = 4 ; Export    // Has a working directory
    LNK_HAS_ARGUMENTS               = 5 ; Export    // Has command line arguments.
    LNK_HAS_CUSTOM_ICON             = 6 ; Export    // Has a custom icon / default icon
        
    // Attribute Properties
    LNK_READONLY                    = 0 ; Export    // Target is read-only
    LNK_HIDDEN                      = 1 ; Export    // Target is hidden
    LNK_SYSTEM                      = 2 ; Export    // Target is a system file
    LNK_VOLUME_LABEL                = 3 ; Export    // Target is a volume label (impossible?)
    LNK_DIRECTORY                   = 4 ; Export    // Target is a directory
    LNK_HAS_BEEN_MODIFIED           = 5 ; Export    // Target has been modified since last backup. (archive)
    LNK_ENCRYPTED                   = 6 ; Export    // Target is encrypted (NTFS EFS)
    LNK_NORMAL                      = 7 ; Export    // Target is "normal"?
    LNK_TEMP                        = 8 ; Export    // Target is temporary
    LNK_SPARSE                      = 9 ; Export    // Target is a sparse file
    LNK_REPARSE_DATA                = 10 ; Export   // Target has reparse point data
    LNK_COMPRESSED                  = 11 ; Export   // Target is compressed
    LNK_OFFLINE                     = 12 ; Export   // Target is offline
    
    // Volume Flags
    LNK_VOLUME_LOCAL                = 0 ; Export    // Available on a local volume
    LNK_VOLUME_NETWORK              = 1 ; Export    // Available on a network share
    
    // Volume Types
    LNK_VOLUME_TYPE_UNKNOWN         = 0 ; Export
    LNK_VOLUME_TYPE_NO_ROOT         = 1 ; Export
    LNK_VOLUME_TYPE_REMOVABLE       = 2 ; Export    // Floppy, zip, etc
    LNK_VOLUME_TYPE_FIXED           = 3 ; Export    // Hard disk
    LNK_VOLUME_TYPE_REMOTE          = 4 ; Export    // Network drive
    LNK_VOLUME_TYPE_OPTICAL         = 5 ; Export    // CD / DVD
    LNK_VOLUME_TYPE_RAM             = 6 ; Export    // RAM Drive

    // ShowWindow Modes (1,2 & 3 are available in the shortcut properties menu, some may be invalid)
    LNK_SWND_HIDE                   = 0 ; Export
    LNK_SWND_NORMAL                 = 1 ; Export
    LNK_SWND_SHOWMINIMIZED          = 2 ; Export
    LNK_SWND_SHOWMAXIMIZED          = 3 ; Export
    LNK_SWND_SHOWNOACTIVATE         = 4 ; Export
    LNK_SWND_SHOW                   = 5 ; Export
    LNK_SWND_MINIMIZE               = 6 ; Export
    LNK_SWND_SHOWMINNOACTIVE        = 7 ; Export
    LNK_SWND_SHOWNA                 = 8 ; Export
    LNK_SWND_RESTORE                = 9 ; Export
    LNK_SWND_SHOWDEFAULT            = 10 ; Export 
    

Type lnks = Record
  lPath             : String = ""
  lTargetPath       : String = ""
  lRelativePath     : String = ""
  lStartIn          : String = ""
  lDescription      : String = ""
  lWorkingDir       : String = ""
  lCustomIcon       : String = ""
  lArguments        : String = "" 
  lIconIndex        : Integer = 0
  lFlag             : Array [7] of Boolean
  lAttribute        : Array [13] of Boolean
  lVolume           : Array [2] of Boolean
  lVolumeType       : Integer = 0
  lVolumeSerial     : Integer = 0
  lVolumeLabel      : String = "UNKNOWN"
  lShareName        : String = ""
  lShowWnd          : Integer = LNK_SWND_NORMAL
  lCreationTime     : Integer = 0
  lModificationTime : Integer = 0
  lLastAccessTime   : Integer = 0
  lTargetFileSize   : Integer = 0
  lHotKey           : Integer = 0
EndType ; Export

Var
    lnk:^lnks ; Export
    lnkList: List of lnks

Function ReadShortcut(lnkpath:String, lnkdebug:Boolean=FALSE) : ^lnks ; Export
{

    Examines a windows .lnk shortcut file, and stores the results.
        
}
Var
    i: Integer
    
    { Reading }
    lnkrfile: Element
    lnkbyte: Integer
    lnkdata: String = ""

    { Lengths }    
    sectlen : Integer
    tblelen : Integer    
    itemlen : Integer

    { Positions }
    pos_start : Integer
    pos_filelocinfo_start : Integer

    { Offsets }
    offset : Integer
    offset_lvoltable : Integer
    offset_lvolbasepath : Integer
    offset_nvoltable : Integer
    offset_filenamepath : Integer
    
Begin

    If Not (Right(lnkpath, 4) = ".lnk") then lnkpath = lnkpath + ".lnk"    
    If lnkdebug then DumpLnk(lnkpath)

    lnk = NewItem(lnkList)
    If Assigned(lnk)
        Result = lnk  
    
        If FileExists(lnkpath) then
            
            { Load to memfile to reduce slow disk reads } 
            lnkrfile = CreateMemFile()
            LoadToMemFile(lnkrfile, lnkpath)
            lnk.lPath = lnkpath
            
            
            { *** HEADER *** }
            
            { The first byte in a .lnk is always 4C, or ASCII for the letter L }
            lnkbyte = ReadByte(lnkrfile)
            If Chr(lnkbyte) = "L" then
            
                { Followed by 3 NULL bytes }                      
                For i = 1 to 3
                    ReadByte(lnkrfile)
                Next
            
                { Followed by a 16 byte GUID (00021401-0000-0000-00C0-000000000046) }
                For i = 1 to 4
                    ReadInt(lnkrfile)
                Next
            
                { Followed by a byte for flags, which indicates what optional
                parts of the file are present...
            
                    bit 6 : LNK_SHELL_PRESENT
                    bit 5 : LNK_POINTS_TO_FILE_OR_DIR
                    bit 4 : LNK_HAS_DESCRIPTION
                    bit 3 : LNK_HAS_RELATIVE_PATH
                    bit 2 : LNK_HAS_WORKING_DIR
                    bit 1 : LNK_HAS_ARGUMENTS
                    bit 0 : LNK_HAS_CUSTOM_ICON

                    bit     0   1  2  3  4 5 6 7                
                    value   128 64 32 16 8 4 2 1
                      
                }
            
                lnkbyte = ReadByte(lnkrfile)

                If Has(lnkbyte,1) then      lnk.lFlag[LNK_SHELL_PRESENT]            = TRUE
                If Has(lnkbyte,2) then      lnk.lFlag[LNK_POINTS_TO_FILE_OR_DIR]    = TRUE
                If Has(lnkbyte,4) then      lnk.lFlag[LNK_HAS_DESCRIPTION]          = TRUE
                If Has(lnkbyte,8) then      lnk.lFlag[LNK_HAS_RELATIVE_PATH]        = TRUE
                If Has(lnkbyte,16) then     lnk.lFlag[LNK_HAS_WORKING_DIR]          = TRUE
                If Has(lnkbyte,32) then     lnk.lFlag[LNK_HAS_ARGUMENTS]            = TRUE
                If Has(lnkbyte,64) then     lnk.lFlag[LNK_HAS_CUSTOM_ICON]          = TRUE            
            
                { Followed by 3 NULL bytes, possibly placeholders for future flags }
                For i = 1 to 3
                    ReadByte(lnkrfile)
                Next
            
                { Followed by two bytes for attributes.
                    If the shortcut target is not a file or directory
                    (LNK_POINTS_TO_FILE_OR_DIR = FALSE), then these bytes are empty.

                    byte 1            
                    bit 7 : LNK_READONLY
                    bit 6 : LNK_HIDDEN
                    bit 5 : LNK_SYSTEM
                    bit 4 : LNK_VOLUME_LABEL
                    bit 3 : LNK_DIRECTORY
                    bit 2 : LNK_HAS_BEEN_MODIFIED
                    bit 1 : LNK_ENCRYPTED
                    bit 0 : LNK_NORMAL
                
                    byte 2
                    bit 7 : LNK_TEMP
                    bit 6 : LNK_SPARSE
                    bit 5 : LNK_REPARSE_DATA
                    bit 4 : LNK_COMPRESSED
                    bit 3 : LNK_OFFLINE

                    bit     0   1  2  3  4 5 6 7                
                    value   128 64 32 16 8 4 2 1
                      
                }
                        
                If lnk.lFlag[LNK_POINTS_TO_FILE_OR_DIR] then     
            
                    lnkbyte = ReadByte(lnkrfile)
                    If Has(lnkbyte,1) then      lnk.lAttribute[LNK_READONLY]             = TRUE
                    If Has(lnkbyte,2) then      lnk.lAttribute[LNK_HIDDEN]               = TRUE
                    If Has(lnkbyte,4) then      lnk.lAttribute[LNK_SYSTEM]               = TRUE
                    If Has(lnkbyte,8) then      lnk.lAttribute[LNK_VOLUME_LABEL]         = TRUE
                    If Has(lnkbyte,16) then     lnk.lAttribute[LNK_DIRECTORY]            = TRUE
                    If Has(lnkbyte,32) then     lnk.lAttribute[LNK_HAS_BEEN_MODIFIED]    = TRUE
                    If Has(lnkbyte,64) then     lnk.lAttribute[LNK_ENCRYPTED]            = TRUE                
                    If Has(lnkbyte,128) then    lnk.lAttribute[LNK_NORMAL]               = TRUE
                
                    lnkbyte = ReadByte(lnkrfile)
                    If Has(lnkbyte,1) then      lnk.lAttribute[LNK_TEMP]                 = TRUE
                    If Has(lnkbyte,2) then      lnk.lAttribute[LNK_SPARSE]               = TRUE
                    If Has(lnkbyte,4) then      lnk.lAttribute[LNK_REPARSE_DATA]         = TRUE
                    If Has(lnkbyte,8) then      lnk.lAttribute[LNK_COMPRESSED]           = TRUE
                    If Has(lnkbyte,16) then     lnk.lAttribute[LNK_OFFLINE]              = TRUE
            
                Else
                    // Skip two bytes...
                    ReadByte(lnkrfile)
                    ReadByte(lnkrfile)
                Endif

                { Followed by 2 NULL bytes, possibly placeholders for future attributes }
                For i = 1 to 2
                    ReadByte(lnkrfile)
                Next
            
                { Followed by three 64 bit integers for the dates (skipping for now) }
                For i = 1 to (3*2)
                    ReadInt(lnkrfile)
                Next
           
                { Followed by an integer which contains the length of the target file in bytes }
                lnk.lTargetFileSize = ReadInt(lnkrfile)
            
                { Followed by an integer which contains the index of the icon to use (zero if no custom icon) }
                lnk.lIconIndex  = ReadInt(lnkrfile)
            
                { Followed by an integer which contains the ShowWnd value to pass }
                lnk.lShowWnd  = ReadInt(lnkrfile)
            
                { Followed by an integer which specifies the hotkey assigned to the shortcut
                    I don't know how to convert this integer into the hotkey,
                    but I assume there's a list somewhere if you're interested
                    in this functionality.                
                }
                lnk.lHotKey  = ReadInt(lnkrfile)
            
                { Followed by eight NULL bytes, probably reserved for the future }
                For i = 1 to 2
                    ReadInt(lnkrfile)
                Next
                
                
                
                { *** BODY *** }
                
                { Shell Item ID List }                
                If lnk.lFlag[LNK_SHELL_PRESENT] then
                    
                    { Short integer containing the Length of the list }
                    sectlen = ReadWord(lnkrfile)
                    
                    { Stub... I don't yet know what information is contained
                        in this list, but by examining the file with a hex
                        editor, you can see that this section does hold
                        a fair ammount of information... I plan to go back
                        to this part.
                    }
                    FileSeek(lnkrfile, FilePosition(lnkrfile) + sectlen)
                
                Endif
                
                
                { File Location Info : Always present, but length may be zero
                    If it's not present, then it's a pretty crap shortcut,
                    because it won't point to anything!!!
                }
                
                    { Integer containing Length of this structure
                        When lnk.lFlag[LNK_POINTS_TO_FILE_OR_DIR] = FALSE, this
                        should also be zero Length. I don't know if I should
                        force this behaviour, or rely on shortcuts to be valid.
                    }
                    pos_filelocinfo_start = FilePosition(lnkrfile)
                    sectlen = ReadInt(lnkrfile)
                    
                    If (sectlen > 0)
                    
                        { Integer containing offset at which this structure ends
                            Should be 1C (28) normally
                        }
                        ReadInt(lnkrfile)
                                           
                        { Followed by a byte for flags, which indicates what type of
                        volumes the file is available on.
            
                            bit 7 : LNK_VOLUME_LOCAL
                            bit 6 : LNK_VOLUME_NETWORK

                            bit     0   1  2  3  4 5 6 7                
                            value   128 64 32 16 8 4 2 1 
                      
                        }
                        
                        lnkbyte = ReadByte(lnkrfile)
                        If Has(lnkbyte,1) then    lnk.lVolume[LNK_VOLUME_LOCAL]   = TRUE
                        If Has(lnkbyte,2) then    lnk.lVolume[LNK_VOLUME_NETWORK] = TRUE
            
                        { Followed by 3 NULL bytes, possibly placeholders for future flags }
                        For i = 1 to 3
                            ReadByte(lnkrfile)
                        Next
                       
                        { Followed by an integer containing the offset to the local volume table
                            This seems to always be 28, and must be relative to the beginning
                            of the file location info structure, i.e. the offset at which
                            this structure ends (the table seems to follow on from this
                            structure) 
                        }
                        offset_lvoltable = ReadInt(lnkrfile)
                        If lnk.lVolume[LNK_VOLUME_LOCAL] = FALSE then offset_lvoltable = 0
                        
                        { Followed by an integer containing the offset to the base path on the local volume }
                        offset_lvolbasepath = ReadInt(lnkrfile)
                        If lnk.lVolume[LNK_VOLUME_LOCAL] = FALSE then offset_lvolbasepath = 0
                        
                        { Followed by an integer containing the offset to the network volume table }
                        offset_nvoltable = ReadInt(lnkrfile)
                        If lnk.lVolume[LNK_VOLUME_NETWORK] = FALSE then offset_nvoltable = 0
                        
                        { Followed by an integer containing the offset to the remaining path of the filename }
                        offset_filenamepath = ReadInt(lnkrfile)
                        
                        { Filename on local volume = basepath + final path
                            Filename on network volume = sharename + final path
                        }
                        
                        
                        { Local Volume Table }
                        If offset_lvoltable > 0 then

                            { Skip to offset
                                The local volume table usually appears after the File Location Info,
                                but I don't know if you can always assume that it will. Therefore,
                                It's easier to seek.
                            }                        
                            FileSeek(lnkrfile, pos_filelocinfo_start + offset_lvoltable)
                        
                            { Followed by an Integer containing the Length of this table,
                                including the variable Length string at the end
                            }
                            pos_start = FilePosition(lnkrfile)
                            tblelen = ReadInt(lnkrfile)
                            
                            { Then an integer containing the type of the volume }
                            lnk.lVolumeType = ReadInt(lnkrfile)
                            
                            { Then an integer containing the volume serial number }
                            lnk.lVolumeSerial = ReadInt(lnkrfile)
                            
                            { Followed by an integer containing an offset to the volume name
                                This should always be 10 (hex), e.g. 16, and is
                                relative to the beginning of the local volume
                                table.
                            }
                            offset = ReadInt(lnkrfile)
                            offset = (pos_start + offset) - FilePosition(lnkrfile)
                            offset = offset 
                            
                            For i = 1 to (offset)
                                ReadByte(lnkrfile)
                            Next
                            
                            { Then just read off the volume label until the end
                                of the structure...
                            }
                            lnk.lVolumeLabel = ""
                            offset = tblelen - (FilePosition(lnkrfile) - pos_start)
                            offset = offset - 1 // NULL terminated 
                            
                            For i = 1 to offset
                                lnk.lVolumeLabel = lnk.lVolumeLabel + Chr(ReadByte(lnkrfile))
                            Next
                            
                            { NULL terminated byte... terminated! }
                            ReadByte(lnkrfile)
                            
                        Endif
                        { End Local Volume Table }  
                        
                        
                        { Network Volume Table }
                        If offset_nvoltable > 0 then
                        
                            { Skip to offset }                            
                            FileSeek(lnkrfile, pos_filelocinfo_start + offset_nvoltable)    
                        
                            { Followed by an Integer containing the Length of this table,
                                including the variable Length string at the end
                            }
                            pos_start = FilePosition(lnkrfile)
                            tblelen = ReadInt(lnkrfile)
                            
                            { Unknown Integer }
                            ReadInt(lnkrfile)
                            
                            { Followed by an integer containing an offset to the share name
                                This should always be 14 (hex), e.g. 20, and is
                                relative to the beginning of the network volume
                                table.
                            }
                            offset = ReadInt(lnkrfile)
                            offset = (pos_start + offset) - FilePosition(lnkrfile)
                            offset = offset 
                            
                            For i = 1 to (offset)
                                ReadByte(lnkrfile)
                            Next
                            
                            { Then just read off the volume label until the end
                                of the structure...
                            }
                            lnk.lShareName = ""
                            offset = tblelen - (FilePosition(lnkrfile) - pos_start)
                            offset = offset - 1 // NULL terminated 
                            
                            For i = 1 to offset
                                lnk.lShareName = lnk.lShareName + Chr(ReadByte(lnkrfile))
                            Next
                            
                            { NULL terminated byte... terminated! }
                            ReadByte(lnkrfile)
                            
                        Endif                       
                        { End Network Volume Table }
                        

                        { Target Path : Local } 
                        If offset_lvoltable > 0 then
                            
                            { Base }
                            FileSeek(lnkrfile, pos_filelocinfo_start + offset_lvolbasepath)
                            lnk.lTargetPath = ReadStringUntilNull(lnkrfile)
                            
                            { Remaining - this usually appears to be emtpy, however
                                searching for this brings us to the correct position
                                to continue reading other data.
                            }
                            FileSeek(lnkrfile, pos_filelocinfo_start + offset_filenamepath )
                            lnk.lTargetPath = lnk.lTargetPath + ReadStringUntilNull(lnkrfile)
                        
                        Endif
                        
                        
                        { Target Path : Network }
                        If offset_nvoltable > 0 then
                        
                            FileSeek(lnkrfile, pos_filelocinfo_start + offset_filenamepath)
                            lnk.lTargetPath = lnk.lShareName + ReadStringUntilNull(lnkrfile)
                            
                        Endif
                        
                    { End File Location Info }
                    
                
                    { Annoyingly, that every string from here on seems to have
                        NULLs every other char, and a Length number that doesn't
                        count these NULLs.
                        
                        Also, the Length number is now stored in a byte, instead of
                        an integer... make up your mind, microsoft!!! :-)
                    }
                    If lnk.lFlag[LNK_HAS_DESCRIPTION] then                        
                        lnk.lDescription = ReadStringStructure(lnkrfile, FALSE)
                    Endif

                    If lnk.lFlag[LNK_HAS_RELATIVE_PATH] then
                        lnk.lRelativePath = ReadStringStructure(lnkrfile, FALSE)                            
                    Endif                         

                    If lnk.lFlag[LNK_HAS_WORKING_DIR] then
                        lnk.lWorkingDir = ReadStringStructure(lnkrfile, FALSE)                            
                    Endif                          

                    If lnk.lFlag[LNK_HAS_ARGUMENTS] then
                        lnk.lArguments = ReadStringStructure(lnkrfile, FALSE)                            
                    Endif

                    If lnk.lFlag[LNK_HAS_CUSTOM_ICON] then
                        lnk.lCustomIcon = ReadStringStructure(lnkrfile, FALSE)                            
                    Endif
                        
                Endif

            Endif
        
            Closefile(lnkrfile)
            
        Endif
    Endif
End

Function CreateShortcut(lnkpath:String, volume: Integer, volume_type: Integer) : ^lnks ; Export
{
    Creates a handle for a new shortcut
}
Begin
    Result = NewItem(lnkList)
    If Assigned(Result)
    
        Result = Result
        
        If Not (Right(lnkpath, 4) = ".lnk") then lnkpath = lnkpath + ".lnk"
        Result.lPath = lnkpath
         
    Endif
End

Function WriteShortcut(lnk: ^lnks) : Boolean ; Export
{
    Writes a shortcut to a file
}
Var
    i : Integer

    { Writing }
    lnkwfile: Element
    lnkbyte: Integer
    lnkdata: String = ""
    
    { Target Attributes }
    fattr : Integer
    target : Element
    targetfilesize : Integer
    
    { Lenghts }
    len_filelocationinfo : Integer
    len_localvolumetable : Integer

Begin
    If FileExists(lnk.lTargetPath, TRUE) then
    
        { Initial Setup : Flags }
        lnk.lFlag[LNK_SHELL_PRESENT] = FALSE // No clue about this, yet
        lnk.lFlag[LNK_POINTS_TO_FILE_OR_DIR] = TRUE // When would a link *not*?
        lnk.lFlag[LNK_HAS_DESCRIPTION] = (Length(lnk.lDescription) > 0)
        lnk.lFlag[LNK_HAS_RELATIVE_PATH] = (Length(lnk.lRelativePath) > 0)
        lnk.lFlag[LNK_HAS_WORKING_DIR] = (Length(lnk.lWorkingDir) > 0)
        lnk.lFlag[LNK_HAS_ARGUMENTS] = (Length(lnk.lArguments) > 0)
        lnk.lFlag[LNK_HAS_CUSTOM_ICON] = (Length(lnk.lCustomIcon) > 0)
        
        { Initial Setup : Attributes }
        fattr = FileAttr(lnk.lTargetPath)
        lnk.lAttribute[LNK_READONLY] = ((fattr And CFA_READONLY) = CFA_READONLY)
        lnk.lAttribute[LNK_HIDDEN] = ((fattr And CFA_HIDDEN) = CFA_HIDDEN)
        lnk.lAttribute[LNK_SYSTEM] = FALSE
        lnk.lAttribute[LNK_VOLUME_LABEL] = FALSE
        lnk.lAttribute[LNK_DIRECTORY] = ((fattr And CFA_DIRECTORY) = CFA_DIRECTORY)
        lnk.lAttribute[LNK_HAS_BEEN_MODIFIED] = TRUE // Usually appears true
        lnk.lAttribute[LNK_ENCRYPTED] = FALSE
        lnk.lAttribute[LNK_NORMAL] = FALSE
        lnk.lAttribute[LNK_TEMP] = FALSE
        lnk.lAttribute[LNK_SPARSE] = FALSE
        lnk.lAttribute[LNK_REPARSE_DATA] = FALSE
        lnk.lAttribute[LNK_COMPRESSED] = FALSE
        lnk.lAttribute[LNK_OFFLINE] = FALSE
        
        { File Size }
        target = ReadFile(lnk.lTargetPath)
        targetfilesize = FileSize(target)
        CloseFile(target)          
            
        { Write to a memfile to reduce slow disk access } 
        lnkwfile = CreateMemFile()

        { First byte is "L" }
        WriteByte(lnkwfile, Asc("L"))
        
        { Three NULL bytes }
        For i = 1 to 3
            WriteByte(lnkwfile, 0)
        Next
        
        { 16 Byte GUID }
        WriteByte(lnkwfile, $01)
        WriteByte(lnkwfile, $14)
        WriteByte(lnkwfile, $02)
        WriteByte(lnkwfile, $00)
        WriteByte(lnkwfile, $00)
        WriteByte(lnkwfile, $00)
        WriteByte(lnkwfile, $00)
        WriteByte(lnkwfile, $00)
        WriteByte(lnkwfile, $C0)
        WriteByte(lnkwfile, $00)
        WriteByte(lnkwfile, $00)
        WriteByte(lnkwfile, $00)
        WriteByte(lnkwfile, $00)
        WriteByte(lnkwfile, $00)
        WriteByte(lnkwfile, $00)
        WriteByte(lnkwfile, $46)
        
        { Followed by a byte for flags, which indicates what optional
        parts of the file are present...
            
            bit 6 : LNK_SHELL_PRESENT
            bit 5 : LNK_POINTS_TO_FILE_OR_DIR
            bit 4 : LNK_HAS_DESCRIPTION
            bit 3 : LNK_HAS_RELATIVE_PATH
            bit 2 : LNK_HAS_WORKING_DIR
            bit 1 : LNK_HAS_ARGUMENTS
            bit 0 : LNK_HAS_CUSTOM_ICON

            bit     0   1  2  3  4 5 6 7                
            value   128 64 32 16 8 4 2 1
                      
        }
            
        lnkbyte = 0

        If lnk.lFlag[LNK_SHELL_PRESENT]         then lnkbyte = lnkbyte + 1
        If lnk.lFlag[LNK_POINTS_TO_FILE_OR_DIR] then lnkbyte = lnkbyte + 2
        If lnk.lFlag[LNK_HAS_DESCRIPTION]       then lnkbyte = lnkbyte + 4
        If lnk.lFlag[LNK_HAS_RELATIVE_PATH]     then lnkbyte = lnkbyte + 8
        If lnk.lFlag[LNK_HAS_WORKING_DIR]       then lnkbyte = lnkbyte + 16
        If lnk.lFlag[LNK_HAS_ARGUMENTS]         then lnkbyte = lnkbyte + 32
        If lnk.lFlag[LNK_HAS_CUSTOM_ICON]       then lnkbyte = lnkbyte + 64
        lnkbyte = lnkbyte + 128 // This always seems to be there - REALLY IMPORTANT!!! 
        WriteByte(lnkwfile, lnkbyte)          
            
        { Followed by 3 NULL bytes, possibly placeholders for future flags }
        For i = 1 to 3
            WriteByte(lnkwfile, 0)
        Next
        
        { Followed by two bytes for attributes.
            If the shortcut target is not a file or directory
            (LNK_POINTS_TO_FILE_OR_DIR = FALSE), then these bytes are empty.

            byte 1            
            bit 7 : LNK_READONLY
            bit 6 : LNK_HIDDEN
            bit 5 : LNK_SYSTEM
            bit 4 : LNK_VOLUME_LABEL
            bit 3 : LNK_DIRECTORY
            bit 2 : LNK_HAS_BEEN_MODIFIED
            bit 1 : LNK_ENCRYPTED
            bit 0 : LNK_NORMAL
                
            byte 2
            bit 7 : LNK_TEMP
            bit 6 : LNK_SPARSE
            bit 5 : LNK_REPARSE_DATA
            bit 4 : LNK_COMPRESSED
            bit 3 : LNK_OFFLINE

            bit     0   1  2  3  4 5 6 7                
            value   128 64 32 16 8 4 2 1
                      
        }
                        
        If lnk.lFlag[LNK_POINTS_TO_FILE_OR_DIR] then    
            
            lnkbyte = 0
            If lnk.lAttribute[LNK_READONLY]             then lnkbyte = lnkbyte + 1
            If lnk.lAttribute[LNK_HIDDEN]               then lnkbyte = lnkbyte + 2
            If lnk.lAttribute[LNK_SYSTEM]               then lnkbyte = lnkbyte + 4
            If lnk.lAttribute[LNK_VOLUME_LABEL]         then lnkbyte = lnkbyte + 8
            If lnk.lAttribute[LNK_DIRECTORY]            then lnkbyte = lnkbyte + 16
            If lnk.lAttribute[LNK_HAS_BEEN_MODIFIED]    then lnkbyte = lnkbyte + 32
            If lnk.lAttribute[LNK_ENCRYPTED]            then lnkbyte = lnkbyte + 64
            If lnk.lAttribute[LNK_NORMAL]               then lnkbyte = lnkbyte + 128
            WriteByte(lnkwfile, lnkbyte)
                
            lnkbyte = 0
            If lnk.lAttribute[LNK_TEMP]                 then lnkbyte = lnkbyte + 1
            If lnk.lAttribute[LNK_SPARSE]               then lnkbyte = lnkbyte + 2
            If lnk.lAttribute[LNK_REPARSE_DATA]         then lnkbyte = lnkbyte + 4
            If lnk.lAttribute[LNK_COMPRESSED]           then lnkbyte = lnkbyte + 8
            If lnk.lAttribute[LNK_OFFLINE]              then lnkbyte = lnkbyte + 16
            lnkbyte = lnkbyte + 32 // This always seems to be there - REALLY IMPORTANT!!!
            WriteByte(lnkwfile, lnkbyte)
            
        Else
            // Skip two bytes...
            WriteByte(lnkwfile, 0)
            WriteByte(lnkwfile, 0)
        Endif
        
        { Followed by 2 NULL bytes, possibly placeholders for future attributes }
        For i = 1 to 2
            WriteByte(lnkwfile, 0)
        Next
            
        { Followed by three 64 bit integers for the dates (skipping for now) }
        For i = 1 to (3*2)
            WriteInt(lnkwfile, 0)
        Next
        
        { Followed by an integer which contains the length of the target file in bytes }       
        WriteInt(lnkwfile, targetfilesize)
            
        { Followed by an integer which contains the index of the icon to use (zero if no custom icon) }
        WriteInt(lnkwfile, lnk.lIconIndex)
            
        { Followed by an integer which contains the ShowWnd value to pass }
        WriteInt(lnkwfile, lnk.lShowWnd)
            
        { Followed by an integer which specifies the hotkey assigned to the shortcut }
        WriteInt(lnkwfile, lnk.lHotKey)
            
        { Followed by eight NULL bytes, probably reserved for the future }
        For i = 1 to 2
            WriteInt(lnkwfile, 0)
        Next
        
        { Shell Item Id List
            To Do...        
        }
        
        { File Location Info }
        
            { The first integer is the total Length of this structure, and all
                following data (including local volume table and the base path) up
                to the description/relative path/etc string.
            }
                        
            len_localvolumetable = 16 + Length(lnk.lVolumeLabel) + 1
            len_filelocationinfo = 28 + len_localvolumetable + Length(lnk.lTargetPath) + 1 + 1
            
            WriteInt(lnkwfile, len_filelocationinfo)
            
            { Pointer to first offset (LVT), should be 1C (28) }
            WriteInt(lnkwfile, 28)
            
            { Followed by a byte for flags, which indicates what type of
                volumes the file is available on.
            
                bit 7 : LNK_VOLUME_LOCAL
                bit 6 : LNK_VOLUME_NETWORK

                bit     0   1  2  3  4 5 6 7                
                value   128 64 32 16 8 4 2 1 
                      
            }
                        
            WriteByte(lnkwfile, 1) // Local
            
            { Followed by 3 NULL bytes, possibly placeholders for future flags }
            For i = 1 to 3
                WriteByte(lnkwfile, 0)
            Next
            
            { Integer offset to the local volume table - same as pointer to the
                first offset, i.e. 1C/28
            }
            WriteInt(lnkwfile, 28)

            { Offset to the base path on the local volume }            
            WriteInt(lnkwfile, 28 + len_localvolumetable)
            
            { Integer offset to the network volume table - 0 in this case }
            WriteInt(lnkwfile, 0)
            
            { Offset to the remaining path }
            WriteInt(lnkwfile, 28 + len_localvolumetable + Length(lnk.lTargetPath) + 1)
                                
            
            { Local Volume Table }
            
                { Length of this structure }
                WriteInt(lnkwfile, len_localvolumetable)
                
                { Volume Type }
                If lnk.lVolumeType = 0 then lnk.lVolumeType = LNK_VOLUME_TYPE_FIXED
                WriteInt(lnkwfile, lnk.lVolumeType)
                
                { Volume Serial Number
                    Does not have to be valid. I'll use zero.
                }
                WriteInt(lnkwfile, 0)
                
                { Offset to the volume label (always 10 hex, i.e. 16) }
                WriteInt(lnkwfile, 16)
                
                { ... Which is a bit useless, because the volume label appears
                straight after this offset! }
                WriteStringAppendNull(lnkwfile, lnk.lVolumeLabel)
            
            { End Local Volume Table }
            
            
            { Target Path }
            WriteStringAppendNull(lnkwfile, lnk.lTargetPath)
            
            { Remaing Path - empty }
            WriteStringAppendNull(lnkwfile, "")
        
        { End File Location Info }
        
        { Strings }
        If lnk.lFlag[LNK_HAS_DESCRIPTION]       then WriteStringStructure(lnkwfile, lnk.lDescription)
        If lnk.lFlag[LNK_HAS_RELATIVE_PATH]     then WriteStringStructure(lnkwfile, lnk.lRelativePath)
        If lnk.lFlag[LNK_HAS_WORKING_DIR]       then WriteStringStructure(lnkwfile, lnk.lWorkingDir)
        If lnk.lFlag[LNK_HAS_ARGUMENTS]         then WriteStringStructure(lnkwfile, lnk.lArguments)
        If lnk.lFlag[LNK_HAS_CUSTOM_ICON]       then WriteStringStructure(lnkwfile, lnk.lCustomIcon)        
        
        { End }
        WriteInt(lnkwfile, 0)

        SaveMemFile(lnkwfile, lnk.lPath)       
        Result = TRUE
    
    Endif
End





Function Has(a: Integer, b: Integer) : Boolean
Begin
    If (a AND b) = b then result = TRUE else result = FALSE
End

Function ReadStringStructure(mfile: Element, cint:Boolean=TRUE) : String
Var
    inbyte : Integer
    itemlen: Integer
    i: Integer
Begin

    { Integer containing Length of the string }                        
    If cint then
        itemlen = ReadInt(mfile)
    Else
        itemlen = ReadByte(mfile) * 2
    Endif
    
    If itemlen > 0 then
    
        For i = 1 to itemlen
        
            inbyte = ReadByte(mfile)
            
            If inbyte > 0 then Result = Result + Chr(inbyte)
                
        Next
        
        { NULL terminated byte }
        ReadByte(mfile)
        
    Else
        Result = ""
    Endif
    
End

Function ReadStringUntilNull(mfile: Element) : String
Var
    inbyte : Integer
Begin

    Result = ""

    inbyte = ReadByte(mfile)
    
    While (inbyte > 0)
        Result = Result + Chr(inbyte)
        inbyte = ReadByte(mfile)
    Wend
    
    
End

Procedure WriteStringAppendNull(mfile: Element, s: String)
Var
    i : Integer
Begin

    If Length(s) > 0 then
        
        For i = 1 to Length(s)
            WriteByte(mfile, Asc(Mid(s,i,1)))
        Next
    
    Endif
    
    WriteByte(mfile, 0) // NULL terminated
    
    
End

Procedure WriteStringStructure(mfile: Element, s: String)
Var
    l, i : Integer
Begin

    l = Length(s)
    WriteByte(mfile, l)
    
    If Length(s) > 0 then
        
        For i = 1 to l
            WriteByte(mfile, 0)
            WriteByte(mfile, Asc(Mid(s,i,1)))
        Next
    
    Endif
    
    WriteByte(mfile, 0) // NULL terminated   
    
End

Procedure DumpLnk(lnkpath:String)
{

    Dump the contents of the lnk file for diagnosis.
    
}
Var
    lnkrfile, lnkwfile: Element
    lnkbyte: Integer
Begin
    
    DeleteFile(lnkpath+".bin")

    lnkrfile = CreateMemFile()
    LoadToMemFile(lnkrfile, lnkpath)
    lnkwfile = WriteFile(lnkpath+".bin")
    While Not EOF(lnkrfile)
        lnkbyte = ReadByte(lnkrfile)
        WriteByte(lnkwfile,lnkbyte)
    Wend
    CloseFile(lnkrfile)
    CloseFile(lnkwfile)
End


Begin
End
